//+------------------------------------------------------------------+
//|                                                   Cnormalize.mqh |
//|                                  Copyright 2024, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2024, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum Enormalize
{  NORMALIZE_NONE = -1,
   NORMALIZE_ROBUST = 0,
   NORMALIZE_FEATURE = 1,
   NORMALIZE_STANDARD = 2,
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class Cnormalize
{
protected:
   void              StandardScore(vector &Data);
   void              FeatureScaling(vector &Data);
   void              RobustScaling(vector &Data);
   void              ReverseFeatureScaling(vector &Data);
   void              ReverseRobustScaling(vector &Data);
public:

   void              Transform(Enormalize Type, vector &Output, vector &BatchTransform, vector &Beta, vector &Gamma);

                     Cnormalize(void) {};
                    ~Cnormalize(void) {};
};
//+------------------------------------------------------------------+
//| Function to apply Standard-Score
//+------------------------------------------------------------------+
void Cnormalize::StandardScore(vector &Data)
{  vector _copy;
   _copy.Copy(Data);
   if(_copy.Std() != 0.0)
   {  for (int i = 0; i < int(Data.Size()); i++)
      {  Data[i] = (_copy[i] - _copy.Mean()) / _copy.Std();
      }
   }
}
//+------------------------------------------------------------------+
//| Function to apply Feature-Scaling
//+------------------------------------------------------------------+
void Cnormalize::FeatureScaling(vector &Data)
{  vector _copy;
   _copy.Copy(Data);
   if(_copy.Max() - _copy.Min() > 0.0)
   {  for (int i = 0; i < int(Data.Size()); i++)
      {  Data[i] = (_copy[i] - _copy.Min()) / (_copy.Max() - _copy.Min());
      }
   }
}
//+------------------------------------------------------------------+
//| Function to apply Robust-Scaling
//+------------------------------------------------------------------+
void Cnormalize::RobustScaling(vector &Data)
{  vector _copy;
   _copy.Copy(Data);
   if(_copy.Percentile(75) - _copy.Percentile(25) > 0.0)
   {  for (int i = 0; i < int(Data.Size()); i++)
      {  Data[i] = (_copy[i] - _copy.Median()) / (_copy.Percentile(75) - _copy.Percentile(25));
      }
   }
}
//+------------------------------------------------------------------+
//| Function to apply Reverse Feature-Scaling
//+------------------------------------------------------------------+
void Cnormalize::ReverseFeatureScaling(vector &Data)
{  vector _copy;
   _copy.Copy(Data);
   for (int i = 0; i < int(Data.Size()); i++)
   {  Data[i] = (_copy[i] * (_copy.Max() - _copy.Min())) + _copy.Min();
   }
}
//+------------------------------------------------------------------+
//| Function to apply Reverse Robust-Scaling
//+------------------------------------------------------------------+
void Cnormalize::ReverseRobustScaling(vector &Data)
{  vector _copy;
   _copy.Copy(Data);
   for (int i = 0; i < int(Data.Size()); i++)
   {  Data[i] = (_copy[i] * (_copy.Percentile(75) - _copy.Percentile(25))) + _copy.Median();
   }
}
//+------------------------------------------------------------------+
//| Batch Normalizing Transform
//+------------------------------------------------------------------+
void Cnormalize::Transform(Enormalize Type, vector &Output, vector &BatchTransform, vector &Beta, vector &Gamma)
{  if(Type != NORMALIZE_NONE)
   {  if(Type == NORMALIZE_STANDARD)
      {  StandardScore(Output);
      }
      else if(Type == NORMALIZE_FEATURE)
      {  FeatureScaling(Output);
      }
      else if(Type == NORMALIZE_ROBUST)
      {  RobustScaling(Output);
      }
      //Transformer
      BatchTransform.Init(Output.Size());
      BatchTransform.Fill(0.0);
      for(int i = 0; i < int(Output.Size()); i++)
      {  BatchTransform[i] = (Gamma[i] * Output[i]) + Beta[i];
      }
   }
}
//+------------------------------------------------------------------+
